/*
 * Copyright (c) 2018 Giam Teck Choon <choon@choon.net> or CHOON.NET.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <bglibs/iobuf.h>
#include <bglibs/str.h>
#include <bglibs/dns.h>

#include "mailfront.h"

#define RWLDOMAINS "/var/qmail/control/rwldomains"
#define RWLDOMAINS6 "/var/qmail/control/rwldomains6"
#define RBLDOMAINS "/var/qmail/control/rbldomains"
#define RBLDOMAINS6 "/var/qmail/control/rbldomains6"

#define eprintf(...) fprintf(stderr, __VA_ARGS__)
#define elogf(A, ...) { if(A){ showpid(); } fprintf(stderr, __VA_ARGS__); }

static response resp;
static const char *mainprog = "mailfront";
static const char *thisprog = "plugin-check-dnsbl";
static pid_t pid;
static const char *linkproto;
static int is_ipv6 = 0;
static int debug = 0;
static const char *local_ip;
static const char *remote_ip;
static const char *relayclient;
static const char *skipcheck;
static const char *rblsmtpd;
static const char *defercheck;
static const char *rejectcheck;
static str rwldomains;
static str rbldomains;
static str estr;
static str hmsg;
static str rblmsg;

void showpid()
{
	if (pid < 0 || !pid) {
		pid = getpid();
	}
	fprintf(stderr, "%s-%s[%d]: ", mainprog, thisprog, pid);
}

static inline char tohex(char c)
{
	return c>=10?c-10+'a':c+'0';
}

int make_query_name(const char *ip, const char *rbl, str *out)
{
	str_init(out);
	// If there is a colon then assume it is IPv6
	if (strchr(ip, ':')) {
		ipv6addr ip6;
		if (!ipv6_scan(ip, &ip6)) {
			if (debug) {
				elogf(1, "ipv6_scan error: (%s:%d) Cannot "
				      "parse IP '%s'\n", __FILE__, __LINE__,
				      ip);
			}
			return 0;
		}
		int j;
		char hexval;
		for (j=16; j>0; j--) {
			hexval = tohex(ip6.addr[j-1] & 15);
			if (!str_catb(out, &hexval, 1)) {
				if (debug) {
					errno = ENOMEM;
					elogf(1, "str_catb error: (%s:%d) "
					      "errno=%d errmsg=%s\n",
					      __FILE__, __LINE__,
					      errno, strerror(errno));
				}
				return 0;
			}
			if (!str_cats(out, ".")) {
				if (debug) {
					errno = ENOMEM;
					elogf(1, "str_cats error: (%s:%d) "
					      "errno=%d errmsg=%s\n",
					      __FILE__, __LINE__,
					      errno, strerror(errno));
				}
				return 0;
			}
			hexval = tohex(ip6.addr[j-1] >> 4);
			if (!str_catb(out, &hexval, 1)) {
				if (debug) {
					errno = ENOMEM;
					elogf(1, "str_catb error: (%s:%d) "
					      "errno=%d errmsg=%s\n",
					      __FILE__, __LINE__,
					      errno, strerror(errno));
				}
				return 0;
			}
			if (!str_cats(out, ".")) {
				if (debug) {
					errno = ENOMEM;
					elogf(1, "str_cats error: (%s:%d) "
					      "errno=%d errmsg=%s\n",
					      __FILE__, __LINE__,
					      errno, strerror(errno));
				}
				return 0;
			}
		}
		if (!str_cats(out, rbl)) {
			if (debug) {
				errno = ENOMEM;
				elogf(1, "str_catb error: (%s:%d) errno=%d "
				      "errmsg=%s\n", __FILE__, __LINE__,
				      errno, strerror(errno));
			}
			return 0;
		}
	}
	else {
		ipv4addr ip4;
		if (!ipv4_scan(ip, &ip4)) {
			if (debug) {
				elogf(1, "ipv4_scan error: (%s:%d) Cannot "
				      "parse IP '%s'\n", __FILE__, __LINE__,
				      ip);
			}
			return 0;
		}
		char iprbuf[16];
		if (!str_copyb(out, iprbuf, fmt_ipv4addr_reverse(iprbuf,
		    &ip4))) {
			if (debug) {
				errno = ENOMEM;
				elogf(1, "str_copyb error: (%s:%d) errno=%d "
				      "errmsg=%s\n", __FILE__, __LINE__,
				      errno, strerror(errno));
			}
			return 0;
		}
		if (!str_cats(out, ".")) {
			if (debug) {
				errno = ENOMEM;
				elogf(1, "str_cats error: (%s:%d) errno=%d "
				      "errmsg=%s\n", __FILE__, __LINE__,
				      errno, strerror(errno));
			}
			return 0;
		}
		if (!str_cats(out, rbl)) {
			if (debug) {
				errno = ENOMEM;
				elogf(1, "str_cats error: (%s:%d) errno=%d "
				      "errmsg=%s\n", __FILE__, __LINE__,
				      errno, strerror(errno));
			}
			return 0;
		}
	}
	return 1;
}

int readfile_str(const char *filename, str *out)
{
	str_init(out);
	ibuf in;
	str line = {0, 0, 0};
	if (!ibuf_open(&in, filename, 0)) return 0;
	while (ibuf_getstr(&in, &line, '\n')) {
		//str_rstrip(&line);
		if (line.len > 0 && line.s[0] != '#') {
			if (!str_cats(out, line.s)) {
				ibuf_close(&in);
				str_free(&line);
				if (debug) {
					errno = ENOMEM;
					elogf(1, "str_cats error: (%s:%d) "
					      "errno=%d errmsg=%s\n",
					      __FILE__, __LINE__, errno,
					      strerror(errno));
				}
				return 0;
			}
		}
	}
	ibuf_close(&in);
	str_free(&line);
	return 1;
}

/* return 1 on success otherwise return 0 */
int init_rwldomains()
{
	const char *tmp;
	str_init(&rwldomains);
	str s = {0, 0, 0};
	if (is_ipv6) {
		tmp = session_getenv("RWLDOMAINS_IPV6");
		if (tmp && *tmp) {
			str_cats(&rwldomains, tmp);
		} else {
			tmp = session_getenv("RWLDOMAINSFILE_IPV6");
			if (tmp && *tmp) {
				if (readfile_str(tmp, &s)) {
					str_cats(&rwldomains, s.s);
				}
			} else {
				if (readfile_str(RWLDOMAINS6, &s)) {
					str_cats(&rwldomains, s.s);
				}
			}
		}
	} else {
		tmp = session_getenv("RWLDOMAINS");
		if (tmp && *tmp) {
			str_cats(&rwldomains, tmp);
		} else {
			tmp = session_getenv("RWLDOMAINSFILE");
			if (tmp && *tmp) {
				if (readfile_str(tmp, &s)) {
					str_cats(&rwldomains, s.s);
				}
			} else {
				if (readfile_str(RWLDOMAINS, &s)) {
					str_cats(&rwldomains, s.s);
				}
			}
		}
	}
	str_free(&s);
	return 1;
}

/* return 1 on success otherwise return 0 */
int init_rbldomains()
{
	const char *tmp;
	str_init(&rbldomains);
	str s = {0, 0, 0};
	if (is_ipv6) {
		tmp = session_getenv("RBLDOMAINS_IPV6");
		if (tmp && *tmp) {
			str_cats(&rbldomains, tmp);
		} else {
			tmp = session_getenv("RBLDOMAINSFILE_IPV6");
			if (tmp && *tmp) {
				if (readfile_str(tmp, &s)) {
					str_cats(&rbldomains, s.s);
				}
			} else {
				if (readfile_str(RBLDOMAINS6, &s)) {
					str_cats(&rbldomains, s.s);
				}
			}
		}
	} else {
		tmp = session_getenv("RBLDOMAINS");
		if (tmp && *tmp) {
			str_cats(&rbldomains, tmp);
		} else {
			tmp = session_getenv("RBLDOMAINSFILE");
			if (tmp && *tmp) {
				if (readfile_str(tmp, &s)) {
					str_cats(&rbldomains, s.s);
				}
			} else {
				if (readfile_str(RBLDOMAINS, &s)) {
					str_cats(&rbldomains, s.s);
				}
			}
		}
	}
	str_free(&s);
	return 1;
}

/*
 * If $RBLSMTPD is set and not empty, first char is - then reply with 553
 * otherwise reply with 451
 * return 1 to block otherwise return 0
 */
int check_rblsmtpd(void)
{
	const char *x = rblsmtpd;
	if (x) {
		if (!*x) return 0;
		if (*x == '-') {
			resp.number = 553;
			++x;
		} else {
			resp.number = 451;
		}
		resp.message = x;
		return 1;
	}
	return 0;
}

/* return 0 on error, 1 on success/not listed and 2 on listed */
int make_txt_query(const char *query, str *out)
{
	str_init(out);
	static struct dns_result txt;
	if (query && *query) {
		if (dns_txt(&txt, query) < 0) {
			if (debug) {
				elogf(1, "DNS error doing TXT lookup for "
				      "'%s'\n", query);
			}
			str_cats(out, "DNS error doing TXT lookup for '");
			str_cats(out, query);
			str_cats(out, "'");
			return 0;
		}
		if (txt.count > 0) {
			if (debug) {
				elogf(1, "DNS TXT result for '%s': '%s'\n",
					 query, txt.rr.name[0]);
			}
			str_cats(out, txt.rr.name[0]);
			return 2;
		}
	}
	return 1;
}

static const response *init(void)
{
	if (pid < 0 || !pid) {
		pid = getpid();
	}
	linkproto = getprotoenv(0);
	is_ipv6 = linkproto != 0 && strcasecmp(linkproto, "TCP6") == 0;
	local_ip = getprotoenv("LOCALIP");
	remote_ip = getprotoenv("REMOTEIP");
	if (remote_ip && *remote_ip) {
		if (strchr(remote_ip, ':')) {
			is_ipv6 = 1;
		}
	}
	relayclient = session_getenv("RELAYCLIENT");
	rejectcheck = session_getenv("DNSBL_CHECK_REJECT");
	defercheck = session_getenv("DNSBL_CHECK_DEFER");
	skipcheck = session_getenv("DNSBL_CHECK_SKIP");
	rblsmtpd = session_getenv("RBLSMTPD");
	if (session_getenv("DNSBL_CHECK_VERBOSE")) {
		debug = 1;
	}
	if (!init_rwldomains() || !init_rbldomains()) {
		elogf(1, "init_rwldomains or init_rbldomains error\n");
		return &resp_internal;
	}
	return 0;
}

static const response *sender(str* address, str* params)
{
	if (skipcheck) {
		if (debug) {
			elogf(1, "Skipped. $DNSBL_CHECK_SKIP is set.\n");
		}
		return 0;
	}
	if (!local_ip) {
		local_ip = getprotoenv("LOCALIP");
	}
	if (!local_ip) {
		elogf(1, "Skipped. $TCPLOCALIP not set.\n");
		return 0;
	}
	if (!remote_ip) {
		remote_ip = getprotoenv("REMOTEIP");
	}
	if (!remote_ip) {
		elogf(1, "Skipped. $TCPREMOTEIP not set.\n");
		return 0;
	}
	str_init(&hmsg);
	/* X-DNSBL-Check-Result: SKIPPED: $RELAYCLIENT is set for 1.2.3.4 */
	if (relayclient) {
		if (debug) {
			elogf(1, "Skipped. $RELAYCLIENT is set for "
			      "$TCPREMOTEIP '%s'.\n", remote_ip);
		}
		str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: "
				"$RELAYCLIENT is set for ");
		str_cats(&hmsg, remote_ip);
		str_cats(&hmsg, "\n");
		return 0;
	}
	/* X-DNSBL-Check-Result: SKIPPED: authenticated user from 1.2.3.4 */
	if (session_getnum("authenticated", 0)) {
		if (debug) {
			elogf(1, "Skipped. Authenticated for $TCPREMOTEIP "
			      "'%s'\n", remote_ip);
		}
		str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: "
				"authenticated user from ip ");
		str_cats(&hmsg, remote_ip);
		str_cats(&hmsg, "\n");
		return 0;
	}
	/*
	 * X-DNSBL-Check-Result: SKIPPED: $RBLSMTPD is set and empty
	 * for 1.1.1.1
	 */
	if (rblsmtpd && rblsmtpd[0] == 0) {
		/* if is set and empty we skip it */
		if (debug) {
			elogf(1, "Skipped. $RBLSMTPD is set and empty for "
			      "$TCPREMOTEIP '%s'\n", remote_ip);
		}
		str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: "
				"$RBLSMTPD is set and empty for ");
		str_cats(&hmsg, remote_ip);
		str_cats(&hmsg, "\n");
		return 0;
	}
	if (check_rblsmtpd()) {
		return &resp;
	}
	/*
	 * X-DNSBL-Check-Result: SKIPPED: $RWLDOMAINS and $RBLDOMAINS not set
	 * or empty for 1.2.3.4
	 */
	if (rwldomains.len == 0 && rbldomains.len == 0) {
		/* if is set and empty we skip it */
		if (is_ipv6) {
			if (debug) {
				elogf(1, "Skipped. $RWLDOMAINS_IPV6 and "
				      "$RBLDOMAINS_IPV6 not set or empty "
				      "for TCPREMOTEIP '%s'\n", remote_ip);
			}
			str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: "
					"$RWLDOMAINS_IPV6 and "
					"$RBLDOMAINS_IPV6 not set or "
					"empty for ");
		} else {
			if (debug) {
				elogf(1, "Skipped. $RWLDOMAINS and "
				      "$RBLDOMAINS not set or empty "
				      "for TCPREMOTEIP '%s'\n", remote_ip);
			}
			str_cats(&hmsg, "X-DNSBL-Check-Result: SKIPPED: "
					"$RWLDOMAINS and $RBLDOMAINS not "
					"set or empty for ");
		}
		str_cats(&hmsg, remote_ip);
		str_cats(&hmsg, "\n");
		return 0;
	}
	int action = 0;
	if (rejectcheck || defercheck) {
		if (rejectcheck) {
			action = 553;
			resp.number = action;
		} else {
			action = 451;
			resp.number = action;
		}
	}
	char *delimiters = " ,\t\n";
	str txtquery = {0, 0, 0};
	str txtrwl = {0, 0, 0};
	str_init(&estr);
	int r = 0;
	int rwl_listed = 0;
	/* rwldomains */
	if (rwldomains.len > 0) {
		char *pch;
		char *tmppch = strdup(rwldomains.s);
		if (tmppch == NULL) {
			if (debug) {
				errno = ENOMEM;
				elogf(1, "(%s:%d) errno=%d errmsg=%s\n",
				      __FILE__, __LINE__, errno,
				      strerror(errno));
			}
			str_free(&txtquery);
			str_free(&txtrwl);
			free(tmppch);
			return &resp_oom;
		}
		pch = strtok(tmppch, delimiters);
		while (pch != NULL) {
			if (*pch != '#' && strlen(pch) > 1) {
				if (make_query_name(remote_ip, pch,
				    &txtquery)) {
					r = make_txt_query(txtquery.s,
							   &txtrwl);
					if (debug) {
						elogf(1, "Checking rwldomain"
						      "='%s' ip='%s' query="
						      "'%s'\n", pch,
						      remote_ip, txtquery.s);
					}
					if (r > 1) {
						elogf(1, "rwldomain='%s' ip="
						      "'%s' query='%s' result"
						      "='%s'\n", pch,
						      remote_ip, txtquery.s,
						      txtrwl.s);
						str_cats(&estr, "X-DNSWL-"
								"Check-"
								"Report: "
								"rwldomain=");
						str_cats(&estr, pch);
						str_cats(&estr, " result=");
						str_cats(&estr, txtrwl.s);
						str_cats(&estr, "\n");
						rwl_listed++;
						break;
					}
				}
			}
			pch = strtok(NULL, delimiters);
		}
		free(tmppch);
	}
	/* rbldomains */
	str txtrbl = {0, 0, 0};
	int rbl_listed = 0;
	str_init(&rblmsg);
	if (rwl_listed == 0 && rbldomains.len > 0) {
		char *pch;
		char *tmppch = strdup(rbldomains.s);
		if (tmppch == NULL) {
			if (debug) {
				errno = ENOMEM;
				elogf(1, "(%s:%d) errno=%d errmsg=%s\n",
				      __FILE__, __LINE__, errno,
				      strerror(errno));
			}
			str_free(&txtquery);
			str_free(&txtrbl);
			free(tmppch);
			return &resp_oom;
		}
		pch = strtok(tmppch, delimiters);
		while (pch != NULL) {
			if (*pch != '#' && strlen(pch) > 1) {
				if (make_query_name(remote_ip, pch,
				    &txtquery)) {
					r = make_txt_query(txtquery.s,
							   &txtrbl);
					if (debug) {
						elogf(1, "Checking rbldomain"
						      "='%s' ip='%s' query="
						      "'%s'\n", pch,
						      remote_ip, txtquery.s);
					}
					if (r > 1) {
						elogf(1, "rbldomain='%s' ip="
						      "'%s' query='%s' result"
						      "='%s'\n", pch,
						      remote_ip, txtquery.s,
						      txtrbl.s);
						str_cats(&estr, "X-DNSBL-"
								"Check-"
								"Report: "
								"rbldomain=");
						str_cats(&estr, pch);
						str_cats(&estr, " result=");
						str_cats(&estr, txtrbl.s);
						str_cats(&estr, "\n");
						rbl_listed++;
						str_cats(&rblmsg, txtrbl.s);
						break;
					}
				}
			}
			pch = strtok(NULL, delimiters);
		}
		free(tmppch);
	}
	str_free(&txtrwl);
	str_free(&txtrbl);
	str_free(&txtquery);
	str_init(&hmsg);
	str_cats(&hmsg, "X-DNSBL-Check-Result: ");
	if (rwl_listed > 0) {
		str_cats(&hmsg, "PASSED for ");
		action = 0;
	} else {
		if (rbl_listed > 0) {
			str_cats(&hmsg, "FAILED for ");
		} else {
			str_cats(&hmsg, "PASSED for ");
			action = 0;
		}
	}
	str_cats(&hmsg, remote_ip);
	str_cats(&hmsg, "\n");
	if (estr.len > 0 && estr.s) {
		str_cats(&hmsg, estr.s);
		str_init(&estr);
	}
	if (action > 0 && rbl_listed > 0 && rblmsg.len > 0 && rblmsg.s) {
		resp.message = rblmsg.s;
		return &resp;
	}
	return 0;
	(void)address;
	(void)params;
}

static const response *data_start(int fd)
{
	if (skipcheck) return 0;
	if (hmsg.len > 0) {
		return backend_data_block(hmsg.s, hmsg.len);
	}
	return 0;
	(void)fd;
}

struct plugin plugin = {
	.version = PLUGIN_VERSION,
	.flags = 0,
	.init = init,
	.sender = sender,
	.data_start = data_start,
};
